import os
import sys

sys.path.insert(0, os.path.abspath(".."))

project = "python-chi"
copyright = "2021, University of Chicago"
author = "Jason Anderson"

version = "0.1"
release = "0.1"

extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]

source_suffix = [".rst"]

master_doc = "index"

exclude_patterns = ["build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_extra_path = ["_extra"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {
#     '**': [
#         'about.html',
#         'navigation.html',
#         'relations.html',
#         'searchbox.html',
#     ],
# }

# Output file base name for HTML help builder.
htmlhelp_basename = "ChameleonCloudPythonAPI"

description = "Chameleon Cloud Python API"

# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "ChameleonCloudPythonAPI.tex", description, "Nick Timkovich", "manual"),
]

# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "chameleoncloudapi", description, [author], 1)]

# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "ChameleonCloudPythonAPI",
        description,
        author,
        "ChameleonCloudPythonAPI",
        "A set of Python abstractions for interfacing with the Chameleon testbed",
        "Miscellaneous",
    ),
]

intersphinx_mapping = {
    "python3": ("https://docs.python.org/3/", None),
    "python37": ("https://docs.python.org/3.7/", None),
    "MySQLdb": ("https://mysqlclient.readthedocs.io/", None),
    "novaclient": ("https://docs.openstack.org/python-novaclient/latest/", None),
}

notebook_examples = [
    ('Making a reservation', 'notebooks/reservations.ipynb', [
        'tests/test_lease.py:example_reserve_node',
        'tests/test_lease.py:example_reserve_floating_ip',
        'tests/test_lease.py:example_reserve_network',
        'tests/test_lease.py:example_reserve_multiple_resources',
    ]),
    ('Launching a bare metal instance', 'notebooks/baremetal.ipynb', [
        'tests/test_server.py:example_create_server',
        'tests/test_server.py:example_wait_for_connectivity',
    ]),
]

nbsphinx_execute = 'never'
# This is processed by Jinja2 and inserted before each notebook
nbsphinx_prolog = r"""
{% set docname = env.doc2path(env.docname, base=None) %}

.. figure:: https://img.shields.io/badge/Chameleon-Open%20Notebook-brightgreen
   :target: https://jupyter.chameleoncloud.org/hub/import?deposition_repo=http&deposition_id=https://python-chi.readthedocs.io/en/latest/{{ docname|e }}&ephemeral=true

"""

import generate_notebook
for title, file, examples in notebook_examples:
    generate_notebook.generate(examples, output_file=file, title=title)
    # Also copy to the extras folder
    generate_notebook.generate(examples, output_file=f'_extras/{file}', title=title)



# DEPRECATED
# This module is an alias to the network module and should not be used.
# It is temporarily left in place so that existing references do not break.
from .network import *

print("""
The 'networking_api_examples' module is deprecated! Use the 'network' module
directly going forward.
""")



import os
import sys
import time

from keystoneauth1.identity.v3 import OidcAccessToken
from keystoneauth1 import loading
from keystoneauth1.loading.conf import _AUTH_SECTION_OPT, _AUTH_TYPE_OPT
from keystoneauth1 import session
from oslo_config import cfg
import requests

from . import jupyterhub

import logging
LOG = logging.getLogger(__name__)

DEFAULT_AUTH_TYPE = 'v3token'
CONF_GROUP = 'chi'
RESOURCE_API_URL = os.getenv('CHI_RESOURCE_API_URL',
                             'https://api.chameleoncloud.org')


def default_key_name():
    username = os.getenv('USER')
    return f'{username}-jupyter' if username else None


session_opts = loading.get_session_conf_options()
adapter_opts = loading.get_adapter_conf_options()
# Settings that do not affect authentication, but are used  to define defaults
# for other operations.
extra_opts = [
    cfg.StrOpt('key-name', default=default_key_name(), help=(
        'Name of the SSH keypair to allow access to launched instances')),
    cfg.StrOpt('image', help=(
        'Name of disk image to use when launching instances')),
    cfg.StrOpt('keypair-private-key', help='Path to the SSH private key file'),
    cfg.StrOpt('keypair-public-key', help='Path to the SSH public key file'),
]
global_options = session_opts + adapter_opts + extra_opts

_auth_plugin = None
_session = None
_sites = {}


def printerr(msg):
    return print(msg, file=sys.stderr)


class SessionWithAccessTokenRefresh(session.Session):
    # How many seconds before expiration should we try to refresh
    REFRESH_THRESHOLD = 60  # seconds

    def __init__(self, auth=None, **kwargs):
        def get_access_token(_auth):
            expires_at = getattr(_auth, '_expires_at', time.time())
            should_refresh = (
                expires_at - time.time() < self.REFRESH_THRESHOLD)
            if not getattr(_auth, '_access_token', None) or should_refresh:
                try:
                    access_token, expires_at = jupyterhub.refresh_access_token()
                    _auth._access_token = access_token
                    _auth._expires_at = expires_at
                except Exception as e:
                    LOG.error('Failed to refresh access_token: %s', e)
            return _auth._access_token

        def set_access_token(_auth, access_token):
            _auth._access_token = access_token

        # Override the access_token property to support dynamically refreshing
        # it if close to expiring. Uses the @property decorator, but called
        # explicitly and overridden on the class (it must be on the class,
        # as Python will bind this as a class descriptor.) It has to also
        # provide a setter so auth.access_token = ... works.
        #
        # TODO(jason): this is definitely a hack :) -- but it's the best
        # thing I could come up with, given how many places the code will need
        # a fresh access token. The auth plugin is hard to override b/c it is
        # being dynamically loaded via an entry point, and overriding on the
        # Session.request function is not enough because there are many other
        # points at which an authentication is performed (e.g., looking up an
        # endpoint.)
        if (auth and isinstance(auth, OidcAccessToken)
            and jupyterhub.is_jupyterhub_env()):
            auth._access_token = None
            auth._expires_at = 0
            # Monkey-patch access_token to be @property dynamic lookup
            auth.__class__.access_token = (
                property(get_access_token, set_access_token))

        super().__init__(auth=auth, **kwargs)


class SessionLoader(loading.session.Session):
    plugin_class = SessionWithAccessTokenRefresh


def _auth_plugins():
    return [
        (name, [o._to_oslo_opt() for o in loader.get_options()])
        for name, loader in loading.get_available_plugin_loaders().items()
        if name.startswith('v3')
    ]


def _default_from_env(opts, group=None):
    def _default(opt):
        all_opts = [opt] + (getattr(opt, 'deprecated_opts', getattr(opt, 'deprecated', None)) or [])
        for o in all_opts:
            v = os.environ.get(f'OS_{o.name.replace("-", "_").upper()}')
            if v:
                return v

    if not group:
        group = CONF_GROUP

    for opt in opts:
        default = _default(opt)
        if default:
            cfg.CONF.set_default(opt.dest, default, group=group)


def _auth_section(auth_plugin):
    return f'{CONF_GROUP}.auth.{auth_plugin}'


def _set_auth_plugin(auth_plugin):
    global _auth_plugin
    auth_section = _auth_section(auth_plugin)
    cfg.CONF.set_override(_AUTH_SECTION_OPT.dest, auth_section, group=CONF_GROUP)
    # Re-register the auth conf options so that the auth_type option is
    # registered in the proper auth_section. An alternative would be to register
    # the auth_type option explicitly within every auth_section we use, but
    # this is an idempotent operation, so we just (ab)use the register function.
    loading.register_auth_conf_options(cfg.CONF, CONF_GROUP)
    cfg.CONF.set_override(_AUTH_TYPE_OPT.dest, auth_plugin, group=auth_section)
    _auth_plugin = auth_plugin


def set(key, value):
    """Set a context parameter by name.

    Args:
        key (str): the parameter name.
        value (any): the parameter value.

    Raises:
        cfg.NoSuchOptError: if the parameter is not supported.
    """
    global _session
    # Special handling for auth_type setting as we have to also tell KSA to
    # start reading auth parameters from the plugin's auth section.
    if key == _AUTH_TYPE_OPT.dest:
        _set_auth_plugin(value)
    elif key in [o.dest for o in global_options]:
        cfg.CONF.set_override(key, value, group=CONF_GROUP)
    else:
        valid_for_some_plugin = False
        # Set for all auth plugins that define this option.
        for auth_plugin, opts in _auth_plugins():
            matches = [o for o in opts if key in [o.name, o.dest]]
            for opt in matches:
                cfg.CONF.set_override(opt.dest, value, group=_auth_section(auth_plugin))
            valid_for_some_plugin |= len(matches) > 0
        if not valid_for_some_plugin:
            raise cfg.NoSuchOptError(key)
    # Invalidate session if setting affects session
    if key not in [o.dest for o in extra_opts]:
        _session = None


def get(key):
    """Get a context parameter by name.

    Args:
        key (str): the parameter name.

    Returns:
        any: the parameter value.

    Raises:
        cfg.NoSuchOptError: if the parameter is not supported.
    """
    global _auth_plugin
    if key in [o.dest for o in global_options]:
        return cfg.CONF[CONF_GROUP][key]
    else:
        return cfg.CONF[_auth_section(_auth_plugin)][key]


def params():
    """List all parameters currently set on the context.

    Returns:
        List[str]: a list of parameter names.
    """
    global _auth_plugin
    keys = list(cfg.CONF[CONF_GROUP].keys())
    keys.extend(list(cfg.CONF[_auth_section(_auth_plugin)].keys()))
    return keys


def use_site(site_name):
    """Configure the global request context to target a particular CHI site.

    Targeting a site will mean that leases, instance launch requests, and any
    other API calls will be sent to that site. By default, no site is selected,
    and one must be explicitly chosen.

    .. code-block:: python

       chi.use_site("CHI@UC")

    Changing the site will affect future calls the client makes, implicitly.
    Therefore something like this is possible:

    .. code-block:: python

       chi.use_site("CHI@UC")
       chi.lease.create_lease("my-uc-lease", reservations)
       chi.use_site("CHI@TACC")
       chi.lease.create_lease("my-tacc-lease", reservations)

    Args:
        site_name (str): The name of the site, e.g., "CHI@UC".
    """
    global _sites
    if not _sites:
        res = requests.get(f'{RESOURCE_API_URL}/sites.json')
        try:
            res.raise_for_status()
            items = res.json().get('items', [])
            _sites = {s['name']: s for s in items}
            if not _sites:
                raise ValueError('No sites returned.')
        except Exception:
            printerr("""Failed to fetch list of available Chameleon sites.
                You can still set the site information manually like this,
                if you know the URL and name:

                    chi.set('auth_url', 'https://chi.uc.chameleoncloud.org:5000/v3')
                    chi.set('region_name', 'CHI@UC')
                """)
            return

    site = _sites.get(site_name)
    if not site:
        raise ValueError((
            f'No site named "{site_name}" exists! Possible values: '
            ', '.join(_sites.keys())))

    # Set important parameters
    set('auth_url', f'{site["web"]}:5000/v3')
    set('region_name', site['name'])

    output = [
        f'Now using {site_name}:',
        f'URL: {site.get("web")}',
        f'Location: {site.get("location")}',
        f'Support contact: {site.get("user_support_contact")}',
    ]
    print('\n'.join(output))


def session():
    """Get a Keystone Session object suitable for authenticating a client.

    Returns:
        keystoneauth1.session.Session: the authentication session object.
    """
    global _session
    if not _session:
        auth = loading.load_auth_from_conf_options(cfg.CONF, CONF_GROUP)
        sess = SessionLoader().load_from_conf_options(cfg.CONF, CONF_GROUP, auth=auth)
        _session = loading.load_adapter_from_conf_options(cfg.CONF, CONF_GROUP, session=sess)
    return _session


def reset():
    """Reset the context, removing all overrides and defaults.

    The ``auth_type`` parameter will be defaulted to the value of the
    OS_AUTH_TYPE environment variable, falling back to "v3token" if not defined.

    All context parameters will revert to the default values inferred from
    environment variables.
    """
    global _session
    global _sites
    _session = None
    _sites = {}
    cfg.CONF.reset()
    _set_auth_plugin(
        os.getenv('OS_AUTH_TYPE',
            os.getenv('OS_AUTH_METHOD', DEFAULT_AUTH_TYPE)))
    _default_from_env(global_options, group=CONF_GROUP)
    for auth_plugin, opts in _auth_plugins():
        _default_from_env(opts, group=_auth_section(auth_plugin))


cfg.CONF.register_group(cfg.OptGroup(CONF_GROUP))
loading.register_auth_conf_options(cfg.CONF, CONF_GROUP)
loading.register_session_conf_options(cfg.CONF, CONF_GROUP)
loading.register_adapter_conf_options(cfg.CONF, CONF_GROUP)
cfg.CONF.register_opts(extra_opts, group=CONF_GROUP)

for auth_plugin, opts in _auth_plugins():
    auth_section = _auth_section(auth_plugin)
    cfg.CONF.register_group(cfg.OptGroup(auth_section))
    cfg.CONF.register_opts(opts, group=auth_section)

reset()



from .clients import connection, blazar, glance, gnocchi, keystone, neutron, nova
from .context import reset, set, get, params, session, use_site


__all__ = [
     'get', 'params', 'reset', 'session', 'set', 'use_site',
    'connection', 'blazar', 'glance', 'gnocchi', 'keystone', 'neutron', 'nova',
]



from .clients import neutron

from neutronclient.common.exceptions import NotFound

import json

__all__ = [
    'get_network',
    'get_network_id',
    'create_network',
    'delete_network',
    'update_network',
    'list_networks',

    'get_subnet',
    'get_subnet_id',
    'create_subnet',
    'delete_subnet',
    'update_subnet',
    'list_subnets',

    'get_port',
    'get_port_id',
    'create_port',
    'update_port',
    'delete_port',
    'list_ports',

    'get_router',
    'get_router_id',
    'create_router',
    'delete_router',
    'update_router',
    'list_routers',

    'add_route_to_router',
    'add_routes_to_router',
    'remove_route_from_router',
    'remove_routes_from_router',
    'remove_all_routes_from_router',
    'add_port_to_router',
    'add_port_to_router_by_name',
    'add_subnet_to_router',
    'add_subnet_to_router_by_name',
    'remove_subnet_from_router',
    'remove_port_from_router',

    'get_free_floating_ip',
    'get_floating_ip',
    'list_floating_ips',
    'bind_floating_ip',

    'nuke_network',
]

PUBLIC_NETWORK = 'public'


def _resolve_id(resource, name) -> str:
    list_fn = getattr(neutron(), f'list_{resource}', None)
    if not callable(list_fn):
        raise ValueError(f'Invalid resource type "{resource}"')
    resources = [
        x for x in list_fn()[resource]
        if x['name'] == name
    ]
    if not resources:
        raise RuntimeError(f'No {resource} found with name {name}')
    elif len(resources) > 1:
        raise RuntimeError(f'Found multiple {resource} with name {name}')
    return resources[0]['id']


def _resolve_resource(resource, name_or_id) -> dict:
    get_fn = getattr(neutron(), f'show_{resource}', None)
    if not callable(get_fn):
        raise ValueError(f'Invalid resource type "{resource}"')
    try:
        res = get_fn(name_or_id)
    except NotFound:
        resource_id = _resolve_id(f'{resource}s', name_or_id)
        res = get_fn(resource_id)
    # Unwrap nested structure
    return res.get(resource)


###########
# Networks
###########

def get_network(ref) -> dict:
    """Get a network by its name or ID.

    Args:
        ref (str): The name or ID of the network.

    Returns:
        The network representation.

    Raises:
        RuntimeError: If the network could not be found, or multiple networks
            were returned for the search term.
    """
    return _resolve_resource('network', ref)


def get_network_id(name) -> str:
    """Look up a network's ID from its name.

    Args:
        name (str): The network name.

    Returns:
        The network's ID, if found.

    Raises:
        RuntimeError: If the network could not be found, or multiple networks
            were returned for the search term.
    """
    return _resolve_id('networks', name)


def create_network(network_name, of_controller_ip=None, of_controller_port=None,
                   vswitch_name=None, provider='physnet1',
                   port_security_enabled=True) -> dict:
    """Create a network.

    For an OpenFlow network include the IP and port of an OpenFlow controller
    on Chameleon or accessible through the public Internet. Include a virtual
    switch name if you plan to add additional private VLANs to this switch.
    Additional VLANs can be connected using a dedicated port corresponding to
    the VLAN tag and can be conrolled using a valid OpenFlow controller.

    Args:
        network_name (str): The new network name.
        of_controller_ip (str): the IP of the optional OpenFlow controller.
            The IP must be accessible on the public Internet.
        of_controller_port (str): the port of the optional OpenFlow controller.
        vswitch_name (str): The virtual switch to use name.
        provider (str): the provider network to use when specifying stitchable
            VLANs (i.e. ExoGENI). Default: 'physnet1'
    """
    desc_parts = []
    if of_controller_ip and of_controller_port:
        desc_parts.append(f'OFController={of_controller_ip}:{of_controller_port}')
    if vswitch_name != None:
        desc_parts.append(f'VSwitchName={vswitch_name}')

    network = neutron().create_network(body={
        'network': {
            'name': network_name,
            'description': ','.join(desc_parts),
            'provider:physical_network': provider,
            'provider:network_type': 'vlan',
            'port_security_enabled': port_security_enabled,
        }
    })

    return network['network']


def delete_network(network_id):
    """Delete the network.

    .. note::
       This does not perform a full teardown of the network, including removing
       subnets and ports. It will only succeed if the network does not have
       any attached entities. See :func:`nuke_network` for a more complete
       teardown function.

    Args:
        network_id (str): The network ID.
    """
    return neutron().delete_network(network_id)


def update_network(network_id):
    raise NotImplementedError()


def list_networks() -> 'list[dict]':
    """List all networks associated with the current project.

    Returns:
        A list of all the found networks.
    """
    return neutron().list_networks()["networks"]


##########
# Subnets
##########

def get_subnet(ref) -> dict:
    """Get a subnet by its name or ID.

    Args:
        ref (str): The name or ID of the subnet.

    Returns:
        The subnet representation.

    Raises:
        RuntimeError: If the subnet could not be found, or multiple subnets
            were returned for the search term.
    """
    return _resolve_resource('subnet', ref)


def get_subnet_id(name) -> str:
    """Look up a subnet's ID from its name.

    Args:
        name (str): The subnet name.

    Returns:
        The subnet's ID, if found.

    Raises:
        RuntimeError: If the subnet could not be found, or multiple subnets
            were returned for the search term.
    """
    return _resolve_id('subnets', name)


def create_subnet(subnet_name, network_id,
                  cidr='192.168.1.0/24',
                  allocation_pool_start=None,
                  allocation_pool_end=None,
                  gateway_ip=None) -> dict:
    """Create a subnet on a network.

    Args:
        subnet_name (str): The name to give the new subnet.
        network_id (str): The network to associate the subnet with ID.
        cidr (str): The subnet's IPv4 CIDR range. (Default 192.168.1.0/24)
        gateway_ip (str): The subnet's gateway address. If not defined,
            the first address in the subnet will be automatically chosen as
            the gateway.

    Returns:
        The new subnet representation.
    """
    subnet = {
        'name': subnet_name,
        'cidr': cidr,
        'ip_version': 4,
        'network_id': network_id,
    }
    if gateway_ip:
        subnet['gateway_ip'] = gateway_ip
    if allocation_pool_start and allocation_pool_end:
        subnet['allocation_pools']= [
            {
            "start": allocation_pool_start,
            "end": allocation_pool_end,
            }
        ]


    subnet_rtn = neutron().create_subnet(body={
        'subnets': [ subnet ]
    })

    return subnet_rtn['subnets'][0]


def delete_subnet(subnet_id):
    """Delete the subnet.

    Args:
        subnet_id (str): The subnet ID.
    """
    return neutron().delete_subnet(subnet_id)


def update_subnet(subnet_id):
    raise NotImplementedError()


def list_subnets() -> 'list[dict]':
    """List all subnets associated with the current project.

    Returns:
        A list of all the found subnets.
    """
    return neutron().list_subnets()["subnets"]


########
# Ports
########

def get_port(ref) -> dict:
    """Get a port by its name or ID.

    Args:
        ref (str): The name or ID of the port.

    Returns:
        The port representation.

    Raises:
        RuntimeError: If the port could not be found, or multiple ports
            were returned for the search term.
    """
    return _resolve_resource('port', ref)


def get_port_id(name) -> str:
    """Look up a port's ID from its name.

    Args:
        name (str): The port name.

    Returns:
        The port's ID, if found.

    Raises:
        RuntimeError: If the port could not be found, or multiple ports
            were returned for the search term.
    """
    return _resolve_id('ports', name)


def create_port(port_name, network_id, fixed_ips=None, subnet_id=None, ip_address=None,
               port_security_enabled=True) -> dict:
    """Create a new port on a network.

    This function has a short-form and a long-form invocation. In the short form,
    you can specify ``subnet_id`` and ``ip_address`` to give the port a single
    assignment on a subnet. In the long form you can specify ``fixed_ips`` to
    define multiple assignments.

    Args:
        port_name (str): The name to give the new port.
        network_id (str): The ID of the network that the port will be
            connected to.
        fixed_ips (list[dict]): A list of IP assignments to give to the port
            on various subnets. Each assignment must at minimum have a ``subnet_id``
            defined. An optional ``ip_address`` can be included on an assignment
            to specify the exact IP address to assign. Otherwise, one is chosen
            automatically from the available IPs on the subnet. There can be
            multiple assignments (i.e., IPs) on a single subnet.
        subnet_id (str): The ID of the subnet that the port will be allocated
            on. The port will be automatically assigned an IP address on this
            subnet, unless the ``ip_address`` parameter is provided.

            .. note::
               This parameter is ignored if ``fixed_ips`` is set.

        ip_address (str): The IP address to assign the port, if a specific
            IP address is desired. By default an IP address is automatically
            picked from the target subnet.

            .. note::
               This parameter is ignored if ``fixed_ips`` is set.

        port_security_enabled (bool): Whether to enable `port security
            <https://wiki.openstack.org/wiki/Neutron/ML2PortSecurityExtensionDriver>`_.
            In general this should be kept on. (Default True).

    Returns:
        The created port representation.
    """
    port = {
        'name': port_name,
        'network_id': network_id,
        'port_security_enabled': port_security_enabled,
    }

    if fixed_ips is None and subnet_id is not None:
        fixed_ip = {'subnet_id': subnet_id}
        if ip_address is not None:
            fixed_ip['ip_address'] = ip_address
        fixed_ips = [fixed_ip]

    port['fixed_ips'] = fixed_ips

    return neutron().create_port(body={'port': port})


def update_port(port_id, subnet_id=None, ip_address=None):
    raise NotImplementedError()


def delete_port(port_id):
    """Delete the port.

    Args:
        port_id (str): The port ID.
    """
    return neutron().delete_port(port_id)


def list_ports() -> 'list[dict]':
    """List all ports associated with the current project.

    Returns:
        A list of all the found ports.
    """
    return neutron().list_ports()["ports"]


##########
# Routers
##########

def get_router(ref) -> dict:
    """Get a router by its name or ID.

    Args:
        ref (str): The name or ID of the router.

    Returns:
        The router representation.

    Raises:
        RuntimeError: If the router could not be found, or multiple routers
            were returned for the search term.
    """
    return _resolve_resource('router', ref)


def get_router_id(name) -> str:
    """Look up a router's ID from its name.

    Args:
        name (str): The router name.

    Returns:
        The router's ID, if found.

    Raises:
        RuntimeError: If the router could not be found, or multiple routers
            were returned for the search term.
    """
    return _resolve_id('routers', name)


def create_router(router_name, gw_network_name=None) -> dict:
    """Create a router, with or without a public gateway.

    Args:
        router_name (str): The new router name.
        gw_network_name (str): The name of the public gateway requested to
            provide subnets connected this router NAT to the Internet.
    Returns:
        The created router representation.
    """
    router = {"name": router_name, "admin_state_up": True}

    if gw_network_name:
        router["external_gateway_info"] = {"network_id": get_network_id(gw_network_name)}

    response = neutron().create_router(body={"router": router})
    return response["router"]


def delete_router(router_id):
    """Delete the router.

    Args:
        router_id (str): The router ID.
    """
    return neutron().delete_router(router_id)


def update_router(router_id):
    raise NotImplementedError()


def list_routers() -> 'list[dict]':
    """List all routers associated with the current project.

    Returns:
        A list of all the found routers.
    """
    return neutron().list_routers()["routers"]


####################
# Router operations
####################

def add_route_to_router(router_id, cidr, nexthop):
    """Add a new route to a router.

    Args:
        router_id (str): The router ID.
        cidr (str): The destination subnet CIDR for the route.
        nexthop (str): The nexthop address for the route.
    """
    return add_routes_to_router(router_id, [
        {'destination': cidr, 'nexthop': nexthop}
    ])


def add_routes_to_router(router_id, routes):
    """Add a set of routes to a router.

    Args:
        router_id (str): The router ID.
        routes (list[dict]): A list of routes to add. The list is expected
            to consist of items with a 'destination' and 'nexthop' key, e.g.:

            .. code-block:: python

                [
                   {'destination': '10.0.0.0/24', 'nexthop': '10.0.0.1'},
                   {'destination': '10.0.1.0/24', 'nexthop': '10.0.1.1'}
                ]

    """
    return neutron().add_extra_routes_to_router(router_id, {
        'router': {'routes': routes}
    })


def remove_route_from_router(router_id, cidr, nexthop):
    """Remove a single route from the router.

    Args:
        router_id (str): The router ID.
        cidr (str): The destination subnet CIDR for the route.
        nexthop (str): The nexthop address for the route.
    """
    return remove_routes_from_router(router_id, [
        {'destination': cidr, 'nexthop': nexthop}
    ])


def remove_routes_from_router(router_id, routes):
    """Remove a set of routes from a router.

    Args:
        router_id (str): The router ID.
        routes (list[dict]): A list of routes to remove. The list is expected
            to consist of items with a 'destination' and 'nexthop' key, e.g.:

            .. code-block:: python

                [
                   {'destination': '10.0.0.0/24', 'nexthop': '10.0.0.1'},
                   {'destination': '10.0.1.0/24', 'nexthop': '10.0.1.1'}
                ]

    """
    return neutron().remove_extra_routes_from_router(router_id, {
        'router': {
            'routes': routes
        }
    })


def remove_all_routes_from_router(router_id):
    """Remove all routes from the router.

    Args:
        router_id (str): The router ID.
    """
    return remove_routes_from_router(router_id,
        get_router(router_id)['routes'])


def add_port_to_router(router_id, port_id):
    """Add a port to a router.

    Args:
        router_id (str): The router ID.
        port_id (str): The port ID.
    """
    return neutron().add_interface_router(router_id, {'port_id': port_id})


def add_port_to_router_by_name(router_name, port_name):
    """Add a port to a router, referencing the router and port by name.

    Args:
        router_name (str): The router name.
        port_name (str): The port name.
    """
    router_id = get_router_id(router_name)
    port_id = get_port_id(port_name)
    return add_port_to_router(router_id, port_id)


def remove_port_from_router(router_id, port_id):
    """Remove a port from the router.

    Args:
        router_id (str): The router ID.
        port_id (str): The port ID.
    """
    return neutron().remove_interface_router(router_id, {'port_id': port_id})


def add_subnet_to_router(router_id, subnet_id):
    """Add a subnet to a router.

    Args:
        router_id (str): The router ID.
        subnet_id (str): The subnet ID.
    """
    return neutron().add_interface_router(router_id, {'subnet_id': subnet_id})


def add_subnet_to_router_by_name(router_name, subnet_name):
    """Add a subnet to a router, referencing the router and subnet by name.

    Args:
        router_name (str): The router name.
        subnet_name (str): The subnet name.
    """
    router_id = get_router_id(router_name)
    subnet_id = get_subnet_id(subnet_name)
    return add_subnet_to_router(router_id, subnet_id)


def remove_subnet_from_router(router_id, subnet_id):
    """Remove a subnet from the router.

    Args:
        router_id (str): The router ID.
        subnet_id (str): The subnet ID.
    """
    return neutron().remove_interface_router(router_id, {
        'subnet_id': subnet_id
    })


###############
# Floating IPs
###############

def get_free_floating_ip(allocate=True) -> dict:
    """Get the first unallocated floating IP available to your project.

    Args:
        allocate (bool): Whether to allocate a new floating IP if there are no
            Floating IPs currently free in your project. Defaults to True.

    Returns:
        The free floating IP representation.
    """
    _neutron = neutron()
    ips = _neutron.list_floatingips()['floatingips']
    unbound = (ip for ip in ips if ip['port_id'] is None)
    try:
        fip = next(unbound)
        return fip
    except StopIteration:
        if not allocate:
            raise RuntimeError(
                "No free floating IPs in project and not allocating a new one")
        return _neutron.create_floatingip({
            "floatingip": {
                "floating_network_id": get_network_id(PUBLIC_NETWORK),
            }
        })["floatingip"]


def get_or_create_floating_ip() -> 'tuple[dict,bool]':
    """Get the first unallocated floating IP or allocate one to the project.

    Returns:
        A tuple of the floating IP representation, and a boolean indicating
            whether the IP was dynamically allocated to the project.

    Raises:
        Conflict: If there are no free floating IPs and there are no more
            available to allocate.
    """
    try:
        # Set allocate=False so we explicitly know if we created one.
        fip = get_free_floating_ip(allocate=False)
        created = False
    except Exception:
        network_id = get_network_id(PUBLIC_NETWORK)
        fip = neutron().create_floatingip({
            'floatingip': {
                'floating_network_id': network_id
            }
        })['floatingip']
        created = True
        print(f'Allocated new floating IP {fip["floating_ip_address"]}')
    return fip, created


def get_floating_ip(ip_address) -> dict:
    """Get the floating IP representation for an IP address.

    Args:
        ip_address (str): The IP address of the floating IP.

    Returns:
        The floating IP representation.
    """
    ips = neutron().list_floatingips()['floatingips']

    for fip in ips:
        if fip['floating_ip_address'] == ip_address:
            return fip
    raise Exception(f"Floating IP {ip_address} not found")


def list_floating_ips() -> 'list[dict]':
    """List all floating ips associated with the current project.

    Returns:
        A list of all the found floating ips.
    """
    return neutron().list_floatingips()["floatingips"]


def bind_floating_ip(ip_address, port_id=None, fixed_ip_address=None):
    """Directly assign a Floating IP to an existing port/address.

    .. note::
       If you just want to attach a Floating IP to a server instance, the
       :func:`chi.server.associate_floating_ip` function is simpler.

    Args:
        ip_address (str): The Floating IP address.
        port_id (str): The ID of the port to bind to.
        fixed_ip_address (str): The address in the port to bind to. This is
            only required if the port has multiple IP addresses assigned; by
            default the first IP in a port is bound.
    """
    fip = get_floating_ip(ip_address)
    neutron().update_floatingip(fip["id"], body={
        "floatingip": {
            "port_id": port_id,
            "fixed_ip_address": fixed_ip_address,
        }
    })


def nuke_network(network_name):
    """Completely tear down the network.

    Cleanly tearing down an OpenStack network representation involves a few
    separate steps:

    1. Detach the network's subnets from the router.
    2. Delete the router.
    3. Delete the subnet(s).
    4. Delete the network.

    This function performs all of those steps for you.

    .. note::

       This function will not work well for very advance networks, perhaps
       those connected to multiple routers. You should perform your own cleanup
       if your network's subnets are attached to multiple routers.

    Args:
        network_name (str): The network name.
    """
    network = get_network(network_name)
    network_id = network["id"]

    #Detach the router from all of its networks
    router_ports = [
        port for port in neutron().list_ports()["ports"]
        if port["device_owner"] == "network:router_interface" and port["network_id"] == network_id
    ]

    for port in router_ports:
        for fixed_ip in port["fixed_ips"]:
            router_id = port["device_id"]
            remove_subnet_from_router(router_id, fixed_ip["subnet_id"])

    #Delete the router
    for port in router_ports:
        delete_router(port["device_id"])

    #Delete the subnet
    for subnet in neutron().list_subnets()['subnets']:
        if subnet['network_id'] == network_id:
            subnet_id=subnet['id']
            delete_subnet(subnet_id)

    #Delete the network
    delete_network(network_id)


###################
# Wizard functions
###################

class wizard(object):
    """A collection of "wizard" functions.

    These utility functions are very opinionated but can reduce boilerplate.
    """

    @staticmethod
    def create_network(name_prefix, of_controller_ip=None,
                       of_controller_port=None, gateway=False):
        """Create a network and subnet, and connect the subnet to a new router.

        Args:
            name_prefix (str): The common name prefix for all created entities.
            of_controller_ip (str): The OpenFlow controller IP, if using.
            of_controller_port (int): The OpenFlow controller port, if using.
            gateway (bool): Whether to add a WAN gateway to the router. Routers
                with a WAN gateway are able to NAT to the Internet.

        Returns:
            The created network representation.
        """
        network_name = f'{name_prefix}Net'
        vswitch_name = f'{name_prefix}VSwitch'
        router_name = f'{name_prefix}Router'
        subnet_name = f'{name_prefix}Subnet'
        network = create_network(
            network_name,
            of_controller_ip=of_controller_ip,
            of_controller_port=of_controller_port,
            vswitch_name=vswitch_name,
            provider='physnet1'
        )
        subnet = create_subnet(subnet_name, network['id'])
        router = create_router(router_name, gateway=gateway)
        add_subnet_to_router(router['id'], subnet['id'])

        return network

    @staticmethod
    def delete_network(name_prefix):
        """Delete a network created via :func:``wizard.create_network``.

        Args:
            name_prefix (str): The common name prefix for all created entities.
        """
        return nuke_network(f'{name_prefix}Net')



#!/usr/bin/env python
import inspect
import json
import os
from textwrap import dedent

import click
import nbformat.v4 as nbf


INTRO_CELLS = [
    nbf.new_markdown_cell(dedent("""
    First, select which project and site you wish to authenticate against.
    """).strip()),
    nbf.new_code_cell(dedent("""
    import chi

    chi.use_site('CHI@UC')
    # Set to your project's charge code
    chi.set('project_name', 'CH-XXXXXX')
    """).strip())
]


def generate_notebook(*example_fns, title=None):
    nb = nbf.new_notebook()
    nb.cells = []
    nb.metadata.language_info = {
        'name': 'Python',
        'version': 3,
    }
    nb.metadata['nbsphinx'] = {'execute': 'never'}

    if title:
        nb.cells.append(nbf.new_markdown_cell(f'# {title}'))

    # Put in the generic intro cells
    nb.cells.extend(INTRO_CELLS)

    for fn in example_fns:
        docs = get_docs(fn)
        source = get_source(fn).replace(docs, '')
        nb.cells.extend([
            nbf.new_markdown_cell(f'## {docs}'),
            nbf.new_code_cell(source),
        ])

    return nb


def write_notebook(nb, file_name):
    print(f'write notebook: {file_name}')
    os.makedirs(os.path.dirname(file_name), exist_ok=True)
    with open(file_name, 'w') as f:
        f.write(json.dumps(nb, indent=2))


def load_function(file_name, example_name):
    import importlib
    module_name = file_name.replace('/', '.').replace('.py', '')
    module = importlib.import_module(module_name)
    return getattr(module, example_name, None)


def get_source(fn):
    source = inspect.getsource(fn)
    if fn.__doc__:
        # This is some pretty kludgy code, but basically it's trying to null
        # out the docstring documentation for the function.
        source = (source.replace(fn.__doc__, '')
            .replace('"""', '')
            .replace(':\n    \n', ':\n'))
    lines = [l for l in source.split('\n') if (not l or l.startswith(' '))]
    return dedent('\n'.join(lines))


def get_docs(fn):
    docstring = inspect.getdoc(fn)
    # Parse "Uses" section
    # print(docstring)
    # print(USES_REGEX.match(docstring))
    return docstring


def generate(examples, output_file=None, title=None):
    fns = []
    for example_ref in examples:
        file_name, fn_name = example_ref.split(':')
        fn = load_function(file_name, fn_name)
        if not fn:
            raise RuntimeError(f'Unable to load {fn_name} from {file_name}')
        fns.append(fn)
    with open(output_file, 'w') as f:
        contents = json.dumps(generate_notebook(*fns, title=title), indent=2)
        f.write(contents + '\n')


@click.command()
@click.option('--output-file', default='notebook.ipynb',
              help='The output notebook file')
@click.option('--title', help='A title to give the Notebook')
@click.argument('examples', nargs=-1)
def cli(examples, output_file=None, title=None):
    """Generate a notebook from a list of example functions.

    EXAMPLES are expected to be a list of files and the functions located
    within those files, for example, to inline the 'reserve_node' function
    located in 'tests/test_lease.py':

      generate_notebook.py tests/test_lease.py:reserve_node

    The outputted notebook will have a Markdown cell with the function's
    docstring, while the source of the function will be included in to a
    following code cell.

    Multiple examples can be provided, in which case the outputted Notebook
    file will have multiple code and markdown cells.
    """
    generate(examples, output_file=output_file, title=title)


if __name__ == '__main__':
    cli()



import os

from oslo_config import cfg
import pytest
import requests_mock

import chi

def setup_function():
    chi.reset()


def test_get():
    value = "KEYNAME"
    chi.set("key_name", value)
    assert chi.get("key_name") == value


def test_get_invalid_key():
    with pytest.raises(cfg.NoSuchOptError):
        chi.get("some_invalid_key")


def test_set():
    values = ["KEYNAME", "KEYNAME2"]
    [chi.set("key_name", k) for k in values]
    assert chi.get("key_name") == values[-1]


def test_set_invalid_key():
    with pytest.raises(cfg.NoSuchOptError):
        chi.set("some_invalid_key", "foo")


def test_reset():
    value = "KEYNAME"
    chi.set("key_name", value)
    chi.reset()
    assert chi.get("key_name") == None


def test_session():
    chi.set("auth_url", "AUTH_URL")
    chi.set("token", "TOKEN")
    chi.set("project_name", "PROJECT_NAME")
    chi.set("project_domain_name", "PROJECT_DOMAIN_NAME")
    chi.set("interface", "INTERFACE")
    chi.set("region_name", "REGION_NAME")
    session = chi.session()
    assert session
    assert session.interface == "INTERFACE"
    assert session.region_name == "REGION_NAME"
    auth_plugin = session.session.auth
    assert auth_plugin
    assert auth_plugin.auth_url == "AUTH_URL"
    assert auth_plugin.auth_methods[0].token == "TOKEN"
    assert auth_plugin.project_name == "PROJECT_NAME"
    assert auth_plugin.project_domain_name == "PROJECT_DOMAIN_NAME"


def test_switch_auth_plugin():
    chi.set("auth_type", "v3oidcaccesstoken")
    chi.set("auth_url", "AUTH_URL")
    chi.set("protocol", "PROTOCOL")
    chi.set("access_token", "ACCESS_TOKEN")
    chi.set("identity_provider", "IDENTITY_PROVIDER")
    session = chi.session()
    assert session
    assert session.session.auth.access_token == "ACCESS_TOKEN"
    chi.set("auth_type", "v3token")
    chi.set("token", "TOKEN")
    session = chi.session()
    assert session
    assert session.session.auth.auth_methods[0].token == "TOKEN"


def test_set_auth_params_in_any_order():
    """It should be possible to set auth_type after setting other parameters.

    Any parameters set before setting auth_type should be preserved and used,
    even if the currently selected auth_type does not use those parameters.
    """
    chi.set("auth_url", "AUTH_URL")
    chi.set("application_credential_id", "APPLICATION_CREDENTIAL_ID")
    chi.set("application_credential_secret", "APPLICATION_CREDENTIAL_SECRET")
    chi.set("auth_type", "v3applicationcredential")
    session = chi.session()
    assert session


def test_default_from_env():
    os.environ['OS_AUTH_TYPE'] = 'v3password'
    os.environ['OS_USERNAME'] = 'USERNAME'
    os.environ['OS_PASSWORD'] = 'PASSWORD'
    os.environ['OS_PROJECT_ID'] = 'PROJECT_ID'
    chi.reset()
    assert chi.get('auth_type') == 'v3password'
    assert chi.get('username') == 'USERNAME'
    assert chi.get('password') == 'PASSWORD'
    assert chi.get('project_id') == 'PROJECT_ID'


def test_use_site():
    with requests_mock.Mocker() as m:
        m.get('https://api.chameleoncloud.org/sites.json', json={'items': [
            {'name': 'foo', 'web': 'http://web', 'user_support_contact': 'help'}
        ]})
        chi.use_site('foo')
        assert chi.get('auth_url') == 'http://web:5000/v3'
        assert chi.get('region_name') == 'foo'


def test_use_site_missing_site():
    with requests_mock.Mocker() as m:
        m.get('https://api.chameleoncloud.org/sites.json', json={'items': [
            {'name': 'foo', 'web': 'http://web', 'user_support_contact': 'help'}
        ]})
        with pytest.raises(ValueError):
            chi.use_site('bar')


def _test_use_site_error_case(**mock_kwargs):
    """Helper function for default error behavior with .use_site"""
    with requests_mock.Mocker() as m:
        chi.set('auth_url', 'before_update')
        chi.set('region_name', 'before_update')
        m.get('https://api.chameleoncloud.org/sites.json', **mock_kwargs)
        chi.use_site('foo')
        assert chi.get('auth_url') == 'before_update'
        assert chi.get('region_name') == 'before_update'


def test_use_site_http_error():
    _test_use_site_error_case(status_code=404)


def test_use_site_empty_list():
    _test_use_site_error_case(json={'items': []})


def test_use_site_malformed():
    _test_use_site_error_case(json=[])



import os
import requests

ACCESS_TOKEN_ENDPOINT = 'tokens'

hub_api_url = os.getenv('JUPYTERHUB_API_URL')
hub_token = os.getenv('JUPYTERHUB_API_TOKEN')


def is_jupyterhub_env():
    return hub_api_url is not None


def call_jupyterhub_api(path, method='GET'):
    res = requests.request(
        url=f'{hub_api_url}/{path}',
        method=method,
        headers={'authorization': f'token {hub_token}'})
    res.raise_for_status()

    return res.json()


def refresh_access_token():
    """Refresh a user's access token via the JupyterHub API.
    This requires a custom handler be installed within JupyterHub; that handler
    is currently a part of the jupyterhub-chameleon PyPI package.
    Returns:
        str: the new access token for the user.
    Raises:
        AuthenticationError: if the access token cannot be refreshed.
    """
    res = call_jupyterhub_api(ACCESS_TOKEN_ENDPOINT)
    access_token = res.get('access_token')
    expires_at = res.get('expires_at')

    if not access_token:
        raise ValueError(f'Failed to get access token: {res}')

    return access_token, expires_at



from .context import session

# Import all of the client classes for type annotations.
# We have to do this because we lazy-import the client definitions
# inside each function to reduce runtime dependencies.
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from openstack.connection import Connection
    from blazarclient.client import Client as BlazarClient
    from glanceclient.client import Client as GlanceClient
    from gnocchiclient.v1.client import Client as GnocchiClient
    from neutronclient.v2_0.client import Client as NeutronClient
    from novaclient.client import Client as NovaClient
    from ironicclient import client as IronicClient
    from keystoneclient.v3.client import Client as KeystoneClient


session_factory = session

NOVA_API_VERSION = "2.10"


def connection(session=None) -> "Connection":
    """Get connection context for OpenStack SDK.

    The returned :class:`openstack.connection.Connection` object has
    several proxy modules attached for each service provided by the cloud.

    .. note::
       For the most part, it is more straightforward to use clients specific
       to the service you are targeting. However, some of the proxy modules
       are useful for operations that span a few services, such as assigning
       a Floating IP to a server instance.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new connection proxy.
    """
    from openstack.config import cloud_region
    from openstack.connection import Connection
    sess = session or session_factory()
    if hasattr(sess, 'session'):
        # Handle Adapters, which have a nested Session
        sess = sess.session
    cloud_config = cloud_region.from_session(sess)
    return Connection(
        config=cloud_config,
        compute_api_version=NOVA_API_VERSION,
    )


def blazar(session=None) -> "BlazarClient":
    """Get a preconfigured client for Blazar, the reservation service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Blazar client.
    """
    from blazarclient.client import Client as BlazarClient
    return BlazarClient('1', service_type='reservation',
        session=(session or session_factory()))


def glance(session=None) -> "GlanceClient":
    """Get a preconfigured client for Glance, the image service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Glance client.
    """
    from glanceclient.client import Client as GlanceClient
    return GlanceClient('2', session=(session or session_factory()))


def gnocchi(session=None) -> "GnocchiClient":
    """Get a preconfigured client for Gnocchi, the metrics service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Gnocchi client.
    """
    from gnocchiclient.v1.client import Client as GnocchiClient
    sess = session or session_factory()
    session_options = dict(auth=sess.session.auth)
    adapter_options = dict(interface=sess.interface, region_name=sess.region_name)
    return GnocchiClient(
        adapter_options=adapter_options, session_options=session_options
    )


def neutron(session=None) -> "NeutronClient":
    """Get a preconfigured client for Neutron, the networking service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Neutron client.
    """
    from neutronclient.v2_0.client import Client as NeutronClient
    return NeutronClient(session=(session or session_factory()))


def nova(session=None) -> "NovaClient":
    """Get a preconfigured client for Nova, the compute service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Nova client.
    """
    from novaclient.client import Client as NovaClient
    return NovaClient(NOVA_API_VERSION, session=(session or session_factory()))


def ironic(session=None) -> "IronicClient":
    """Get a preconfigured client for Ironic, the bare metal service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Ironic client.
    """
    from ironicclient import client as IronicClient
    return IronicClient.get_client(
        '1',
        session=session,
        region_name=getattr(session, 'region_name', None),
        # Ironic client defaults to 1.9 currently,
        # "latest" will be latest the API supports
        os_ironic_api_version='latest'
    )


def keystone(session=None) -> "KeystoneClient":
    """Get a preconfigured client for Keystone, the authentication service.

    Args:
        session (Session): An authentication session object. By default a
            new session is created via :func:`chi.session`.

    Returns:
        A new Keystone client.
    """
    from keystoneclient.v3.client import Client as KeystoneClient
    sess = session or session_factory()
    # We have to set interface/region_name also on the Keystone client, as it
    # does not smartly inherit the value sent in on a KSA Adapter instance.
    return KeystoneClient(session=sess,
        interface=getattr(sess, 'interface', None),
        region_name=getattr(sess, 'region_name', None))



import base64
from datetime import datetime
from dateutil import tz
from hashlib import md5
import os


def random_base32(n_bytes):
    rand_bytes = os.urandom(n_bytes)
    return base64.b32encode(rand_bytes).decode("ascii").strip("=")


def sshkey_fingerprint(public_key):
    # See: https://stackoverflow.com/a/6682934
    key = base64.b64decode(public_key.strip().split()[1].encode("ascii"))
    fp_plain = md5(key).hexdigest()
    return ':'.join(a + b for a, b in zip(fp_plain[::2], fp_plain[1::2]))


def get_public_network(neutronclient):
    nets = neutronclient.list_networks()["networks"]
    for net in nets:
        if net["router:external"] != True:
            continue
        pubnet_id = net["id"]
        break
    else:
        raise RuntimeError("couldn't find public net")
    return pubnet_id


def utcnow():
    return datetime.now(tz=tz.tzutc())



from datetime import datetime
from operator import attrgetter
import socket
import time

from novaclient.v2.flavor_access import FlavorAccess as NovaFlavor
from novaclient.v2.keypairs import Keypair as NovaKeypair
from novaclient.v2.servers import Server as NovaServer
from novaclient.exceptions import NotFound

from .clients import connection, glance, nova, neutron
from .context import get as get_from_context, session
from .keypair import Keypair
from .image import get_image, get_image_id
from .network import (get_network_id, get_or_create_floating_ip,
                      get_floating_ip, get_free_floating_ip)
from .util import random_base32, sshkey_fingerprint

__all__ = [
    'get_flavor',
    'get_flavor_id',
    'show_flavor',
    'show_flavor_by_name',
    'list_flavors',

    'get_server',
    'get_server_id',
    'list_servers',
    'delete_server',
    'show_server',
    'show_server_by_name',

    'create_server',

    'associate_floating_ip',
    'detach_floating_ip',

    'wait_for_active',
    'wait_for_tcp',

    'update_keypair',
]

DEFAULT_IMAGE = 'CC-CentOS7'
DEFAULT_NETWORK = 'sharednet1'
BAREMETAL_FLAVOR = 'baremetal'

class ServerError(RuntimeError):
    def __init__(self, msg, server):
        super().__init__(msg)
        self.server = server


def instance_create_args(
    reservation,
    name=None,
    image=DEFAULT_IMAGE,
    flavor=None,
    key=None,
    net_ids=None,
    **kwargs
):
    if name is None:
        name = "instance-{}".format(random_base32(6))

    server_args = {
        "name": name,
        "flavor": flavor,
        "image": image,
        "scheduler_hints": {"reservation": reservation,},
        "key_name": key,
    }

    if net_ids is None:
        # automatically binds "the one" unless there's more than one
        server_args["nics"] = None
    else:
        # Not sure what fields are actually required and what they're called.
        # novaclient (and Nova HTTP API) docs seem vague. the command at
        # https://github.com/ChameleonCloud/horizon/blob/stable/liberty/openstack_dashboard/dashboards/project/instances/workflows/create_instance.py#L943
        # appears to POST a JSON akin to
        # {"server": {..., "networks": [{"uuid": "e8c33574-5423-436c-a45b-5bab78071b8a"}] ...}, "os:scheduler_hints": ...},
        server_args["nics"] = [
            {"net-id": netid, "v4-fixed-ip": ""} for netid in net_ids
        ]

    server_args.update(kwargs)
    return server_args


class Server(object):
    """A wrapper object referring to a server instance.

    This class is helpful if you want to use a more object-oriented programming
    approach when building your infrastrucutre. With the Server abstraction,
    you can for example do the following:

    .. code-block:: python

        with Server(lease=my_lease, image=my_image) as server:
            # When entering this block, the server is guaranteed to be
            # in the "ACTIVE" state if it launched successfully.
            server.associate_floating_ip()
            # Interact with the server (via, e.g., SSH), then...
        # When the block exits, the server will be terminated and deleted

    The above example uses a context manager. The class can also be used
    without a context manager:

    .. code-block:: python

        # Triggers the launch of the server instance
        server = Server(lease=my_lease, image=my_image)
        # Wait for server to be active
        server.wait()
        server.associate_floating_ip()
        # Interact with the server, then...
        server.delete()

    Attributes:
        id (str): The ID of an existing server instance. Use this if you have
            already launched the instance and just want a convenient wrapper
            object for it.
        lease (Lease): The Lease the instance will be launched under.
        key (str): The name of the key pair to associate with the image. This
            is only applicable if launching the image; key pairs cannot be
            added to a server that has already been launched and wrapped via
            the ``id`` attribute.
        image (str): The name or ID of the disk iage to use.
        name (str): A name to give the new instance. (Defaults to an
            auto-generated name.)
        net_ids (list[str]): A list of network IDs to associate the instance
            with. The instance will obtain an IP address on each network
            during boot.

            .. note::
               For bare metal instances, the number of network IDs cannot
               exceed the number of enabled NICs on the bare metal node.

        kwargs: Additional keyword arguments to pass to Nova's server
            :meth:`~novaclient.v2.servers.ServerManager.create` function.
    """
    def __init__(self, id=None, lease=None, key=None, image=DEFAULT_IMAGE,
                 **kwargs):
        kwargs.setdefault("session", session())
        self.session = kwargs.pop("session")
        self.conn = connection(session=self.session)
        self.neutron = neutron(session=self.session)
        self.nova = nova(session=self.session)
        self.glance = glance(session=self.session)
        self.image = get_image(image)
        self.flavor = show_flavor_by_name(BAREMETAL_FLAVOR)

        self.ip = None
        self._fip = None
        self._fip_created = False
        self._preexisting = False

        kwargs.setdefault("_no_clean", False)
        self._noclean = kwargs.pop("_no_clean")

        net_ids = kwargs.pop("net_ids", None)
        net_name = kwargs.pop("net_name", DEFAULT_NETWORK)
        if net_ids is None and net_name is not None:
            net_ids = [get_network_id(net_name)]

        if id is not None:
            self._preexisting = True
            self.server = self.nova.servers.get(id)
        elif lease is not None:
            if key is None:
                key = Keypair().key_name
            server_kwargs = instance_create_args(
                lease.node_reservation,
                image=self.image,
                flavor=self.flavor,
                key=key,
                net_ids=net_ids,
                **kwargs
            )
            self.server = self.nova.servers.create(**server_kwargs)
        else:
            raise ValueError("Missing required argument: 'id' or 'lease' required.")

        self.id = self.server.id
        self.name = self.server.name

    def __repr__(self):
        return "<{} '{}' ({})>".format(self.__class__.__name__, self.name, self.id)

    def __enter__(self):
        self.wait()
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        if exc is not None and self._noclean:
            print("Instance existing uncleanly (noclean = True).")
            return

        self.disassociate_floating_ip()
        if not self._preexisting:
            self.delete()

    def refresh(self):
        """Poll the latest state of the server instance."""
        now = datetime.now()
        try:
            lr = self._last_refresh
        except AttributeError:
            pass  # expected failure on first pass
        else:
            # limit refreshes to once/sec.
            if (now - lr).total_seconds() < 1:
                return

        self.server = self.conn.compute.get_server(self.server)
        self._last_refresh = now

    @property
    def status(self) -> str:
        """Get the instance status."""
        self.refresh()
        return self.server.status

    @property
    def ready(self) -> bool:
        """Check if the instance is marked as active."""
        return self.status == "ACTIVE"

    @property
    def error(self) -> bool:
        """Check if the instance is in an error state."""
        return self.status == "ERROR"

    def wait(self):
        """Wait for the server instance to finish launching.

        If the server goes into an error state, this function will return early.
        """
        self.conn.compute.wait_for_server(self.server, wait=(60 * 20))

    def associate_floating_ip(self):
        """Attach a floating IP to this server instance."""
        if self.ip is not None:
            return self.ip

        self._fip, self._fip_created = get_or_create_floating_ip()
        self.ip = self._fip["floating_ip_address"]
        self.conn.compute.add_floating_ip_to_server(self.server.id, self.ip)
        return self.ip

    def disassociate_floating_ip(self):
        """Detach the floating IP attached to this server instance, if any."""
        if self.ip is None:
            return

        self.conn.compute.remove_floating_ip_from_server(self.server.id, self.ip)
        if self._fip_created:
            self.neutron.delete_floatingip(self._fip["id"])

        self.ip = None
        self._fip = None
        self._fip_created = False

    def delete(self):
        """Delete this server instance."""
        self.conn.compute.delete_server(self.server)
        self.conn.compute.wait_for_server(self.server, status='DELETED')

    def rebuild(self, image_ref):
        """Rebuild this server instance.

        .. note::
           For bare metal instances, this effectively redeploys to the host and
           overwrites the local disk.
        """
        self.image = get_image(image_ref)
        self.conn.compute.rebuild_server(self.server, image=self.image.id)


##########
# Flavors
##########

def get_flavor(ref) -> NovaFlavor:
    """Get a flavor by its ID or name.

    Args:
        ref (str): The ID or name of the flavor.

    Returns:
        The flavor matching the ID or name.

    Raises:
        NotFound: If the flavor could not be found.
    """
    try:
        return show_flavor(ref)
    except NotFound:
        return show_flavor(get_flavor_id(ref))


def get_flavor_id(name) -> str:
    """Look up a flavor's ID from its name.

    Args:
        name (str): The name of the flavor.

    Returns:
        The ID of the found flavor.

    Raises:
        NotFound: If the flavor could not be found.
    """
    flavor = next((f for f in nova().flavors.list() if f.name == name), None)
    if not flavor:
        raise NotFound(f'No flavors found matching name {name}')
    return flavor


def show_flavor(flavor_id) -> NovaFlavor:
    """Get a flavor by its ID.

    Args:
        flavor_id (str): the ID of the flavor

    Returns:
        The flavor with the given ID.
    """
    return nova().flavors.get(flavor_id)


def show_flavor_by_name(name) -> NovaFlavor:
    """Get a flavor by its name.

    Args:
        name (str): The name of the flavor.

    Returns:
        The flavor with the given name.

    Raises:
        NotFound: If the flavor could not be found.
    """
    flavor_id = get_flavor_id(name)
    return show_flavor(flavor_id)


def list_flavors() -> 'list[NovaFlavor]':
    """Get a list of all available flavors.

    Returns:
        A list of all flavors.
    """
    return nova().flavors.list()


##########
# Servers
##########

def get_server(ref) -> NovaServer:
    """Get a server by its ID.

    Args:
        ref (str): The ID or name of the server.

    Returns:
        The server matching the ID.

    Raises:
        NotFound: If the server could not be found.
    """
    try:
        return show_server(ref)
    except NotFound:
        return show_server(get_server_id(ref))


def get_server_id(name) -> str:
    """Look up a server's ID from its name.

    Args:
        name (str): The name of the server.

    Returns:
        The ID of the found server.

    Raises:
        NotFound: If the server could not be found.
    """
    servers = [s for s in nova().servers.list() if s.name == name]
    if not servers:
        raise ValueError(f'No matching servers found for name "{name}"')
    elif len(servers) > 1:
        raise ValueError(f'Multiple matching servers found for name "{name}"')
    return servers[0].id


def list_servers(**kwargs) -> "list[NovaServer]":
    """List all servers under the current project.

    Args:
        kwargs: Keyword arguments, which will be passed to
            :func:`novaclient.v2.servers.list`. For example, to filter by
            instance name, provide ``search_opts={'name': 'my-instance'}``

    Returns:
        All servers associated with the current project.
    """
    return nova().servers.list(**kwargs)


def delete_server(server_id):
    """Delete a server by its ID.

    Args:
        server_id (str): The ID of the server to delete.
    """
    return nova().servers.delete(server_id)


def show_server(server_id) -> NovaServer:
    """Get a server by its ID.

    Args:
        server_id (str): the ID of the server

    Returns:
        The server with the given ID.
    """
    return nova().servers.get(server_id)


def show_server_by_name(name) -> NovaServer:
    """Get a server by its name.

    Args:
        name (str): The name of the server.

    Returns:
        The server with the given name.

    Raises:
        NotFound: If the server could not be found.
    """
    server_id = get_server_id(name)
    return show_server(server_id)


def associate_floating_ip(server_id, floating_ip_address=None):
    """Associate an allocated Floating IP with a server.

    If no Floating IP is specified, one will be allocated dynamically.

    Args:
        server_id (str): The ID of the server.
        floating_ip_address (str): The IPv4 address of the Floating IP to
            assign. If specified, this Floating IP must already be allocated
            to the project.

    """
    if not floating_ip_address:
        floating_ip_address = get_free_floating_ip()["floating_ip_address"]

    connection().compute.add_floating_ip_to_server(server_id, floating_ip_address)

    return floating_ip_address


def detach_floating_ip(server_id, floating_ip_address):
    """Remove an allocated Floating IP from a server by name.

    Args:
        server_id (str): The name of the server.
        floating_ip_address (str): The IPv4 address of the Floating IP to
            remove from the server.

    """
    connection().compute.remove_floating_ip_from_server(
        server_id, floating_ip_address)


def wait_for_active(server_id, timeout=(60 * 20)):
    """Wait for the server to go in to the ACTIVE state.

    If the server goes in to an ERROR state, this function will terminate. This
    is a blocking function.

    .. note::

       For bare metal servers, when the server transitions to ACTIVE state, this
       actually indicates it has started its final boot. It may still take some
       time for the boot to complete and interfaces e.g., SSH to come up.

       If you want to wait for a TCP service like SSH, refer to
       :func:`wait_for_tcp`.

    Args:
        server_id (str): The ID of the server.
        timeout (int): The number of seconds to wait for before giving up.
            Defaults to 20 minutes.

    """
    from openstack.compute.v2.server import Server
    compute = connection().compute
    server = compute._get_resource(Server, server_id)
    return compute.wait_for_server(server, wait=timeout)


def wait_for_tcp(host, port, timeout=(60 * 20)):
    """Wait until a port on a server starts accepting TCP connections.

    The implementation is taken from `wait_for_tcp_port.py
    <https://gist.github.com/butla/2d9a4c0f35ea47b7452156c96a4e7b12>`_.

    Args:
        host (str): The host that should be accepting connections. This can
            be either a Floating IP or a hostname.
        port (int): Port number.
        timeout (int): How long to wait before raising errors, in seconds.
            Defaults to 20 minutes.

    Raises:
        TimeoutError: If the port isn't accepting connection after time
            specified in `timeout`.
    """
    start_time = time.perf_counter()

    while True:
        try:
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as ex:
            time.sleep(0.01)
            if time.perf_counter() - start_time >= timeout:
                raise TimeoutError((
                    f'Waited too long for the port {port} on host {host} to '
                    'start accepting connections.')) from ex

############
# Key pairs
############

def update_keypair(key_name=None, public_key=None) -> "NovaKeypair":
    """Update a key pair's public key.

    Due to how OpenStack Nova works, this requires deleting and re-creating the
    key even for public key updates. The key will not be re-created if it
    already exists and the fingerprints match.

    Args:
        key_name (str): The name of the key pair to update. Defaults to value
            of the "key_name" context variable.
        public_key (str): The public key to update the key pair to reference.
            Defaults to the contents of the file specified by the
            "keypair_public_key" context variable.

    Returns:
        The updated (or created) key pair.
    """
    if not key_name:
        key_name = get_from_context("key_name")
    if not public_key:
        public_key_path = get_from_context("keypair_public_key")
        if public_key_path:
            with open(public_key_path, "r") as pubkey:
                public_key = pubkey.read().strip()

    assert key_name is not None
    assert public_key is not None

    _nova = nova()
    try:
        existing = _nova.keypairs.get(key_name)
        if existing.fingerprint == sshkey_fingerprint(public_key):
            return existing
        _nova.keypairs.delete(key_name)
        return _nova.keypairs.create(
            key_name, public_key=public_key, key_type="ssh")
    except NotFound:
        return _nova.keypairs.create(
            key_name, public_key=public_key, key_type="ssh")


##########
# Wizards
##########

def create_server(server_name, reservation_id=None, key_name=None, network_id=None,
                  network_name=DEFAULT_NETWORK, nics=[], image_id=None,
                  image_name=DEFAULT_IMAGE, flavor_id=None,
                  flavor_name=None, count=1) -> 'Union[NovaServer,list[NovaServer]]':
    """Launch a new server instance.

    Args:
        server_name (str): A name to give the server.
        reservation_id (str): The ID of the Blazar reservation that will be
            used to select a target host node. It is required to make a
            reservation for bare metal server instances.
        key_name (str): A key pair name to associate with the server. Any user
            holding the private key for the key pair will be able to SSH to
            the instance as the ``cc`` user. Defaults to the key specified
            by the "key_name" context variable.
        network_id (str): The network ID to connect the server to. The server
            will obtain an IP address on this network when it boots.
        network_name (str): The name of the network to connect the server to.
            If ``network_id`` is also set, that takes priority.
        nics (list[dict]): ...
        image_id (str): The image ID to use for the server's disk image.
        image_name (str): The name of the image to user for the server's disk
            image. If ``image_id`` is also set, that takes priority.
            (Default ``DEFAULT_IMAGE``.)
        flavor_id (str): The flavor ID to use when launching the server. If not
            set, and no ``flavor_name`` is set, the first flavor found is used.
        flavor_name (str): The name of the flavor to use when launching the
            server. If ``flavor_id`` is also set, that takes priority. If not
            set, and no ``flavor_id`` is set, the first flavor found is used.
        count (int): The number of instances to launch. When launching bare
            metal server instances, this number must be less than or equal to
            the total number of hosts reserved. (Default 1).

    Returns:
        The created server instance. If ``count`` was larger than 1, then a
            list of all created instances will be returned instead.

    Raises:
        ValueError: if an invalid count is provided.
    """
    if count < 1:
        raise ValueError('Must launch at least one server.')
    if not key_name:
        key_name = update_keypair().id
    if not network_id:
        network_id = get_network_id(network_name)
    if not nics:
        nics = [{'net-id': network_id, 'v4-fixed-ip': ''}]
    if not image_id:
        image_id = get_image_id(image_name)
    if not flavor_id:
        if flavor_name:
            flavor_id = get_flavor_id(flavor_name)
        else:
            flavor_id = next((f.id for f in list_flavors()), None)
            if not flavor_id:
                raise NotFound('Could not auto-select flavor to use')

    scheduler_hints = {}
    if reservation_id:
        scheduler_hints['reservation'] = reservation_id

    server = nova().servers.create(
        name=server_name,
        image=image_id,
        flavor=flavor_id,
        scheduler_hints=scheduler_hints,
        key_name=key_name,
        nics=nics,
        min_count=count,
        max_count=count
    )
    if count > 1:
        matching = list_servers(search_opts={'name': f'{server_name}-'})
        # In case there are others matching the name, just get the latest
        # batch of instances.
        return sorted(matching, key=attrgetter('created'), reverse=True)[:count]
    else:
        return server



from chi.ssh import Remote

LOCALHOST = "127.0.0.1"


def test_get():
    r = Remote(ip=LOCALHOST)
    assert r.host == LOCALHOST
    assert r.user == "cc"


def test_get_from_server():
    class FakeServer:
        def __init__(self):
            self.ip = LOCALHOST

    r = Remote(server=FakeServer())
    assert r.host == LOCALHOST
    assert r.user == "cc"
